import inspect
from collections.abc import Callable


def pretty_func(func: Callable, /) -> str:
    func = inspect.unwrap(func)
    func_name: str = _get_name(func)
    return f"{func_name}()"


def _get_name(func: Callable) -> str:
    return (
        getattr(func, "__qualname__", None)
        or getattr(func, "__name__", None)
        or "<unknown>"
    )



from pathlib import Path

import pydantic
import pydantic_settings

from ._base import BaseConfig
from ._joblib import ConfigJoblib
from ._log_level import LogLevel
from ._paths import paths
from ._pretty import ConfigPretty


class Config(BaseConfig):
    model_config = pydantic_settings.SettingsConfigDict(env_prefix="LIBLAF_GRAPES_")

    joblib: ConfigJoblib = pydantic.Field(default_factory=ConfigJoblib)
    log_file: Path = pydantic.Field(default=paths.log_file)
    log_level: LogLevel = pydantic.Field(default=LogLevel.INFO)
    pretty: ConfigPretty = pydantic.Field(default_factory=ConfigPretty)


config = Config()



import datetime
import functools
from collections.abc import Callable
from typing import Any, Protocol, TypedDict, overload

import joblib

from liblaf.grapes._config import config

from ._wrapt import decorator


class ReduceSizeKwargs(TypedDict, total=False):
    bytes_limit: int | str | None
    items_limit: int | None
    age_limit: datetime.timedelta | None


class MemorizedFunc[**P, T](Protocol):
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T: ...


@overload
def cache[**P, T](
    func: Callable[P, T],
    /,
    *,
    memory: joblib.Memory | None = None,
    reduce_size: ReduceSizeKwargs | None = None,
    **kwargs: Any,
) -> MemorizedFunc[P, T]: ...
@overload
def cache[**P, T](
    *,
    memory: joblib.Memory | None = None,
    reduce_size: ReduceSizeKwargs | None = None,
    **kwargs: Any,
) -> Callable[[Callable[P, T]], MemorizedFunc[P, T]]: ...
def cache(
    func: Callable | None = None,
    /,
    *,
    memory: joblib.Memory | None = None,
    reduce_size: ReduceSizeKwargs | None = None,
    **kwargs: Any,
) -> Any:
    if func is None:
        return functools.partial(
            cache, memory=memory, reduce_size=reduce_size, **kwargs
        )
    if memory is None:
        memory = joblib.Memory(config.joblib.memory.location)
    if reduce_size is None:
        reduce_size = {"bytes_limit": config.joblib.memory.bytes_limit}
    func: MemorizedFunc = memory.cache(func, **kwargs)  # pyright: ignore[reportAssignmentType]

    @decorator
    def wrapper(
        wrapped: Callable, _instance: Any, args: tuple, kwargs: dict[str, Any]
    ) -> Any:
        ret: Any = wrapped(*args, **kwargs)
        memory.reduce_size(**reduce_size)
        return ret

    proxy: MemorizedFunc = wrapper(func)
    proxy._self_memory = memory  # pyright: ignore[reportAttributeAccessIssue] # noqa: SLF001
    return proxy



from collections.abc import Mapping
from typing import Any

import pydantic

from ._base import BaseModel


class ConfigPretty(BaseModel):
    """.

    References:
        1. <https://docs.kidger.site/wadler_lindig/api/#wadler_lindig.pformat>
    """

    width: int | None = pydantic.Field(default=None)
    indent: int = pydantic.Field(default=2)
    short_arrays: bool = pydantic.Field(default=True)
    hide_defaults: bool = pydantic.Field(default=True)
    show_type_module: bool = pydantic.Field(default=True)
    show_dataclass_module: bool = pydantic.Field(default=False)
    show_function_module: bool = pydantic.Field(default=False)
    respect_pdoc: bool = pydantic.Field(default=True)

    @property
    def kwargs(self) -> Mapping[str, Any]:
        return self.model_dump(mode="python", exclude_unset=True, exclude_none=True)



from collections.abc import Mapping, Sequence


def merge[KT, VT](
    *mappings: Mapping[KT, VT], append_arrays: bool = False
) -> dict[KT, VT]:
    result: dict[KT, VT] = {}
    for mapping in mappings:
        for key, value in mapping.items():
            if key in result:
                if isinstance(result[key], Mapping):
                    result[key] = merge(  # pyright: ignore[reportArgumentType]
                        result[key],  # pyright: ignore[reportArgumentType]
                        value,  # pyright: ignore[reportArgumentType]
                        append_arrays=append_arrays,
                    )
                    continue
                if append_arrays and isinstance(result[key], Sequence):
                    result[key] = [*result[key], *value]  # pyright: ignore[reportArgumentType, reportGeneralTypeIssues]
                    continue
            result[key] = value
    return result



import functools
from typing import Any

import attrs
import cytoolz as toolz
import wadler_lindig as wl

from liblaf.grapes._config import config

from ._console import get_console

UNINITIALIZED = wl.TextDoc("<uninitialized>")


def pdoc_attrs(self: Any, **kwargs) -> wl.AbstractDoc:
    """.

    References:
        1. <https://github.com/patrick-kidger/wadler_lindig/blob/0226340d56f0c18e10cd4d375cf7ea25818359b8/wadler_lindig/_definitions.py#L308-L326>
    """
    kwargs: dict[str, Any] = toolz.merge(config.pretty.kwargs, kwargs)
    cls: type = type(self)
    objs: list[tuple[str, Any]] = []
    for field in attrs.fields(cls):
        field: attrs.Attribute
        if not field.repr:
            continue
        value: Any = getattr(self, field.name, UNINITIALIZED)
        if kwargs["hide_defaults"] and value is field.default:
            continue
        objs.append((field.name, value))
    name_kwargs: dict[str, Any] = toolz.assoc(
        kwargs, "show_type_module", kwargs["show_dataclass_module"]
    )
    return wl.bracketed(
        begin=wl.pdoc(cls, **name_kwargs) + wl.TextDoc("("),
        docs=wl.named_objs(objs, **kwargs),
        sep=wl.comma,
        end=wl.TextDoc(")"),
        indent=kwargs["indent"],
    )


@functools.singledispatch
def pformat(obj: Any, **kwargs) -> str:
    kwargs: dict[str, Any] = toolz.merge(config.pretty.kwargs, kwargs)
    if "width" not in kwargs:
        kwargs["width"] = get_console(stderr=True).width
    if not hasattr(obj, "__pdoc__") and attrs.has(type(obj)):
        return pformat_attrs(obj, **kwargs)
    return wl.pformat(obj, **kwargs)


def pformat_attrs(obj: Any, **kwargs) -> str:
    kwargs: dict[str, Any] = toolz.merge(config.pretty.kwargs, kwargs)
    return wl.pformat(pdoc_attrs(obj, **kwargs), **kwargs)


def wadler_lindig[T](cls: type[T]) -> type[T]:
    cls.__repr__ = _repr
    if not hasattr(cls, "__pdoc__") and attrs.has(cls):
        cls.__pdoc__ = pdoc_attrs  # pyright: ignore[reportAttributeAccessIssue]
    return cls


def _repr(self: Any) -> str:
    return pformat(self)
